/*
 *  This code is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 *   
 *  This code is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *   
 *  You should have received a copy of the GNU General Public License
 *  along with GNU Make; see the file COPYING.  If not, write to
 *  the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. 
*/

#include "Cell.h"

#include <memory.h>

#ifdef WIN32
// Functions to read and write values into buffers, taking care of little-endian issues.
void PutInt(uchar* Buffer, int Offset, uint Value)
{
	// DVDs are stored big-endian, so we have to do some work here.
	Buffer += Offset;
	Buffer[3] = (uchar) (Value & 0xFF); Value >>= 8;
	Buffer[2] = (uchar) (Value & 0xFF); Value >>= 8;
	Buffer[1] = (uchar) (Value & 0xFF); Value >>= 8;
	Buffer[0] = (uchar) (Value & 0xFF);
}

void PutShort(uchar* Buffer, int Offset, ushort Value)
{
	Buffer += Offset;
	Buffer[1] = (uchar) (Value & 0xFF); Value >>= 8;
	Buffer[0] = (uchar) (Value & 0xFF); Value >>= 8;
}

uint GetInt(uchar* Buffer, int Offset)
{
	Buffer += Offset;
	uint Ret = 	(Buffer[0] << 24) + (Buffer[1] << 16) + (Buffer[2] << 8) + Buffer[3];
	return Ret;
}

ushort GetShort(uchar* Buffer, int Offset)
{
	Buffer += Offset;
	ushort Ret = (Buffer[0] << 8) + Buffer[1];
	return Ret;
}

// These ones simply swap bytes...
uint FixEndian(uint Input)
{
	return GetInt((uchar*) &Input, 0);
}

ushort FixEndian(ushort Input)
{
	return GetShort((uchar*) &Input, 0);
}

void AddShort(uchar* Buffer, int Offset, short IncValue)
{
	ushort Val = GetShort(Buffer,Offset) + IncValue;
	PutShort(Buffer,Offset,Val);
}

void AddInt(uchar* Buffer, int Offset, int IncValue)
{
	uint Val = GetInt(Buffer,Offset) + IncValue;
	PutInt(Buffer,Offset,Val);
}


#else
// Functions to read and write values into buffers, taking care of little-endian issues.
void PutInt(uchar* Buffer, int Offset, uint Value)
{
	*((uint*)(Buffer+Offset)) = Value;
}

void PutShort(uchar* Buffer, int Offset, ushort Value)
{
	*((ushort*)(Buffer+Offset)) = Value;
}

uint GetInt(uchar* Buffer, int Offset)
{
	return *((uint*)(Buffer+Offset));
}

ushort GetShort(uchar* Buffer, int Offset)
{
	return *((short*)(Buffer+Offset));
}

uint FixEndian(uint Input)
{
	return Input;
}

ushort FixEndian(ushort Input)
{
	return Input;
}

void AddShort(uchar* Buffer, int Offset, short IncValue)
{
	*((ushort*)(Buffer+Offset)) += Value;
}

void AddInt(uchar* Buffer, int Offset, int IncValue)
{
	*((uint*)(Buffer+Offset)) = Value;
}

#endif


#if 0
typedef struct
{
	uint	m_Gard1		: 1;
	uint	m_SRCFrac	: 9;
	uint	m_Gard2		: 1;
	uint	m_SRCInt1	: 15;
	uint	m_Gard3		: 1;
	uint	m_SRCInt2	: 5;
} FullSRC1;

typedef struct
{
	ushort	m_SRCInt2	: 10;
	ushort	m_Gard1		: 1;
	ushort	m_SRCInt3	: 2;
	ushort	m_SRCTop	: 1;
	ushort	m_Gard2		: 2;
} FullSRC2;
#endif

#define DEF_NUM_PACKS 524288	// Enough for a full 1GB file.

VFormat	VideoFormat = FORMAT_UNKNOWN;

CCell::CCell(int NumPacks)
{
	Initialize(NumPacks);
}

CCell::CCell(FILE* Fd)
{
	if(Fd == NULL)
	{
		m_NumAllocPacks = DEF_NUM_PACKS ;
	}
	else
	{
		fseek(Fd,0,SEEK_END);
		m_NumAllocPacks = ftell(Fd) / SECTOR_SIZE;
		fseek(Fd,0,SEEK_SET);
	}
	Initialize(m_NumAllocPacks);
}

void CCell::Initialize(int NumPacks)
{
	m_StartLBA = 0;
	m_EndLBA = 0;
	m_VOBID = 0;		
	m_CELLID = 0;		
	m_StartPTS = 0;		
	m_EndPTS = 0;		
	m_NavPackCount = 0;	
	m_KeepAudio = 0;
	m_AdjustSRC = 0;
	m_FoundButtons = 0;
	m_IFrameCount = 0;
	m_FirstButtonIndex = -1;
	m_LastButtonChange = -1;
	memset(m_HeaderData,0,10);
	m_VOBUTable = NULL;
	m_FixSubPicStops = true;
	m_PTSFirstNavPack = 0;
	m_LBAOfFistButtonNavPack = -1;
	m_MinButtonSize = 5;
	m_MaxSizeX = 0;
	m_MaxSizeY = 0;
	m_AllNops = 0;
	m_ForcedAction = 0;
	m_AnyLinkNext = 0;
	m_PreviewAllCells = 0;

	m_NumPacks = 0;	
	if(NumPacks <= 0)
		m_NumAllocPacks = DEF_NUM_PACKS;
	else
		m_NumAllocPacks = NumPacks;
	m_Packs = new CPack[m_NumAllocPacks];
	memset(m_NumberSubpicUnits,0,32*sizeof(int));

	m_NumFoundAudioStreams = 0;
	memset(m_FoundAudioStreams,0,8*sizeof(uint));

	m_InputImage = 0;
}


CCell::~CCell()
{
	delete[] m_Packs;
	delete[] m_VOBUTable ;
	if(m_InputImage)
		fclose(m_InputImage);
}

//FILE* foo = fopen("/temp/cells.txt","w");

int CCell::ReadFromFile(FILE* InputFile, int &CumulReadLBA, bool IgnoreVideo)
{
	m_NumPacks = 0;
	m_VOBID = 0;
	m_CELLID = 0;

	ushort PackVOBID;
	uchar PackCELLID;

	// Find out the start LBA of this cell.
	m_StartLBA = ftell(InputFile) / SECTOR_SIZE;
	int LBACount = m_StartLBA;
	int EndOfIframe = 0;
	int i;
	CPack* ThisPack;

	short* IFrameSize = 0;		// Size of I frame.

	m_IFrameCount = 0;	
	m_NavPackCount = 0;
	m_FoundButtons = 0;
	m_FirstButtonIndex = -1;
	m_LastButtonChange = -1;

	memset(m_Packs, 0, m_NumAllocPacks*sizeof(CPack));
	memset(m_NumberSubpicUnits,0,32*sizeof(int));
	
	for(i=0; i <= m_NumAllocPacks; i++,LBACount++,CumulReadLBA++)
	{
		// Read next sector.
		if(fread(m_Buf,1,SECTOR_SIZE,InputFile) != SECTOR_SIZE)
		{
			// End of file!
			m_EndLBA = -1;
			break;
		}

		if(i == m_NumAllocPacks)
		{
			// NO MORE SPACE! ERROR!
			ERROR1("The cell is too big to fit in the CCell class! exiting...\n");
			ERRORF1("La cellule est trop grand pour la classe CCell! Annulation...\n");
			return -1;
		}

		// Find pack type.
		ThisPack = &m_Packs[i];
		ThisPack->FindPackType(m_Buf,PackVOBID,PackCELLID);
		ThisPack->m_InputLBA = LBACount;

		if(ThisPack->m_PackType == PackTypeBlank)
		{
			MESSAGE2("Found blank pack at LBA %d\n",LBACount);
			MESSAGEF2("Paquet vide trouv au secteur %d\n",LBACount);
		}
		// JEAN: KLUDGE!
		if(ThisPack->m_PackType == PackTypeSupPic)
		{
			// I'm using ThisPack->m_NumButtons to store whether this pack is the start first pack of a
			// SPU (it has PTS info). This will be useful later on.
			if(m_Buf[0x15] & 0x80)
			{
				ThisPack->m_NumButtons = 1;
				// Also increment our SPU counter for that stream.
				m_NumberSubpicUnits[ThisPack->m_StreamID-0x20]++;
			}
		}
		if(ThisPack->m_PackType == PackTypeAudio)
		{
			// Keep track of which audio streams are present. 
			int St;
			for(St=0;St<m_NumFoundAudioStreams;St++)
			{
				if(ThisPack->m_StreamID == m_FoundAudioStreams[St])
					break;
			}
			if(St == m_NumFoundAudioStreams)
			{
				assert(St < 8);
				m_FoundAudioStreams[m_NumFoundAudioStreams++] = ThisPack->m_StreamID;
			}
		}
		if(ThisPack->m_PackType == PackTypeNavPack)
		{
			// Found a NAVPACK. 
			if(m_VOBID == 0)
			{
				m_VOBID = PackVOBID;
				m_CELLID = PackCELLID;
			}

			// Let's get the start PTS from the first NavPack.
			if(i==0)
				m_StartPTS = GetInt(m_Buf,0x39);

			// Let's see if we're still in the same cell!
			if(m_VOBID != PackVOBID || m_CELLID != PackCELLID)
			{
				m_EndLBA = LBACount;
				break;
			}

			// When the cell is done reading, this will contain the end PTS of the last VOBU...
			m_EndPTS = GetInt(m_Buf,0x3d);
			
			// Locate the end of the I-frame.
			EndOfIframe = GetInt(m_Buf,0x413)+LBACount;
			m_NavPackCount++;

			// Remember if we found any buttons in any navpack...
			if(ThisPack->m_NumButtons > 0)
				m_FoundButtons = true;
			if(m_FirstButtonIndex == -1 && ThisPack->m_NumButtons > 0)
			{
				m_FirstButtonIndex = m_LastButtonChange = i;
				ThisPack->FindMaxButtonSize(m_Buf,m_MaxSizeX,m_MaxSizeY,m_AllNops,m_ForcedAction,m_AnyLinkNext);
			}
			else if(ThisPack->m_NumButtons > 0)
			{
				// This is not the first navpack with buttons. Let's see if the status is 2 as it normally should
				ushort Status = GetShort(m_Buf,0x8d);
				if(Status != 2)
				{
					m_LastButtonChange = i;
					ThisPack->FindMaxButtonSize(m_Buf,m_MaxSizeX,m_MaxSizeY,m_AllNops,m_ForcedAction,m_AnyLinkNext);
				}
			}
		}
		if(ThisPack->m_PackType == PackTypeVideo && !IgnoreVideo)
		{
			// VIDEO PACK
			// Find type of video frame.
			int Offset = 0;
			ThisPack->FindFrameType(m_Buf,&Offset);
#if 0
if(m_VOBID == 1 && m_CELLID == 1)
{
fprintf(foo,"%d %d\n",ThisPack->m_FrameType,ThisPack->m_InputLBA);
fflush(foo);
}
#endif
			// Now, until we get to the end of the I-Frame, we'll say that all the frames are I.
			if(EndOfIframe != -1)
			{
				// It's possible to have 2 I-Frames in a row! ("A new hope, VTS2 5/1). To not count
				// the second one as a new I-Frame, I have to check LBACount
				if(ThisPack->m_FrameType == FrameTypeI && LBACount != EndOfIframe)
				{
					// This frame is an I-Frame.
					ThisPack->m_IFrameType = IFrameTypeStart;
					ushort PESHeaderDataContentFlag = GetShort(m_Buf,0x14);
					IFrameSize = &ThisPack->m_IFrameSize;
					// Keep track of where to write the I-Frame size. A bit risky...
					*IFrameSize = 1;
				}
				else if(LBACount != EndOfIframe)
				{
					// Continuation of I-frame:
					ThisPack->m_FrameType = FrameTypeI;
					ThisPack->m_IFrameType = IFrameTypeMiddle;
					if(IFrameSize)
						*IFrameSize += 1;
#if 0
for(int u=0;u<0x17;u++)
printf("0x%x, ",m_Buf[u]);
#else
#endif
				}
				else
				{
					// end of I-frame:
					// Let's check that it's a frame we can use: we have space for our end_of_sequence.
					Offset = ThisPack->FindEndFrameOffset(m_Buf);
					// We need 4 bytes for our end-of-sequence.
					if(SECTOR_SIZE - Offset < 4)
					{
						// That's a problem, not enough space. Let's flag the Iframe as too full.
						ThisPack->m_IFrameTooFull = 1;
						// If this isn't true, then I need to add some 0's to the end of the I-frame. FIX
						assert(Offset == SECTOR_SIZE - 3);
					}
					// Update size of I-Frame.
					if(IFrameSize)
						*IFrameSize += 1;
					ThisPack->m_FrameType = FrameTypeI;
					ThisPack->m_IFrameType = IFrameTypeEnd;
					IFrameSize = 0;
					m_IFrameCount++;
					EndOfIframe = -1;
				}
			}
			else
			{
				// We're after the end of the I-Frame. All other frames will be called P...
				ThisPack->m_FrameType = FrameTypeP;
			}
		}
	}

	m_NumPacks = i;
	// DONE! Let's seek back so we're at the beginning of the next cell.
	fseek(InputFile,m_EndLBA*SECTOR_SIZE,SEEK_SET);

	return 0;
}


int CCell::ReadNextCellFromFile(FILE* InputFile, int &CumulReadLBA, bool IgnoreVideo)
{
	// Go to m_EndLBA!
	if(m_EndLBA == -1)
		return -1;
	int ret = fseek(InputFile,m_EndLBA*SECTOR_SIZE,SEEK_SET);
	if(ret != 0)
	{
		ERROR2("Can't seek to LBA %d\n",m_EndLBA);
		ERRORF2("Impossible d'aller au secteur %d\n",m_EndLBA);
		return -1;
	}

	return ReadFromFile(InputFile, CumulReadLBA, IgnoreVideo);
}

int CCell::FindThisCell(FILE* InputFile, ushort VOBID, uchar CELLID)
{

	fseek(InputFile,0,SEEK_SET);
	CPack TempPack;
	CPack* ThisPack = &TempPack;
	int InLBACount = 0;
	int ret;
	ushort PackVOBID;
	uchar PackCELLID;

	if(fread(m_Buf,1,SECTOR_SIZE,InputFile) != SECTOR_SIZE)
	{
		// End of file!
		return -1;
	}

	while(1)
	{
		// Is this a nav pack?
		ThisPack->FindPackType(m_Buf,PackVOBID,PackCELLID);
		if(ThisPack->m_PackType != PackTypeNavPack)
		{
			// LOST! ERROR!
			ERROR2("I got lost navigating nav pack at LBA %d! You need a mock strip! exiting! \n",InLBACount);
			ERRORF2("Les pointeurs de navpacks sont errons au secteur %d! Un mock strip est ncessaire! Annulation \n",InLBACount);
			return -1;
		}

		if(VOBID == PackVOBID && CELLID == PackCELLID)
		{
			// We found it!
			fseek(InputFile,InLBACount*SECTOR_SIZE,SEEK_SET);
			return InLBACount;
		}
		if(m_VOBID != PackVOBID || m_CELLID != PackCELLID)
		{
			MESSAGE4("Found VOB/CELL ID %d/%d (LBA %d)\n",PackVOBID,PackCELLID,
				InLBACount);
		}
		m_VOBID = PackVOBID;
		m_CELLID = PackCELLID;


		// Well, we're not there yet. Let's find the next navpack.
		long NextNavLBA = GetInt(m_Buf,0x40f);
		InLBACount += NextNavLBA + 1;
		// Seek to next nav pack.
		ret = fseek(InputFile,InLBACount*SECTOR_SIZE,SEEK_SET);
		if(ret != 0)
		{
			// ERROR!
			ERROR2("Couldn't seek to sector %d, exiting \n",InLBACount);
			ERRORF2("Impossible d'aller au secteur %d, annulation\n",InLBACount);
			return -1;
		}
		// Read one LBA.
		ret = fread(m_Buf,sizeof(uchar),SECTOR_SIZE,InputFile);
		if(ret != SECTOR_SIZE)
		{
			return -1;
		}
	}
	return -1;
}

int CCell::WritePack(FILE* OutputFile, int PackIndex)
{
	// Write 1 LBA:
	if(m_Packs[PackIndex].m_PackType == PackTypeNavPack)
	{
		// Let's at least adjust the two this->lba pointers!
		uint CurrentLBA = ftell(OutputFile)/SECTOR_SIZE;
		PutInt(m_Buf,0x2d,CurrentLBA);
		PutInt(m_Buf,0x40b,CurrentLBA);
	}
	int ret = fwrite(m_Buf,sizeof(uchar),SECTOR_SIZE,OutputFile);
	if(ret != SECTOR_SIZE)
	{
		ERROR2("Couldn't write sector %d, exiting \n",m_StartLBA + PackIndex);
		ERRORF2("Impossible d'crire au secteur %d, anulation\n",m_StartLBA + PackIndex);
		return -1;
	}
	return 0;
}

int CCell::WriteToFile(FILE* InputFile, FILE* OutputFile, int AdjustPacks )
{
	int i;

	m_StartLBA = ftell(OutputFile)/SECTOR_SIZE;

	for(i=0;i<m_NumPacks;i++)
	{
		if(LoadPack(InputFile,i) != 0)
		{
			ERROR1("Read failed, exiting\n");
			ERRORF1("Impossible de lire. Annulation\n");
			return -1;
		}
		if(AdjustPacks)
		{
			if(AdjustAndWritePack(OutputFile,i) != 0)
			{
				ERROR1("Write failed, exiting\n");
				ERRORF1("Impossible d'crire, annulation\n");
				return -1;
			}
		}
		else
		{
			if(WritePack(OutputFile,i) != 0)
			{
				ERROR1("Write failed, exiting\n");
				ERRORF1("Impossible d'crire, annulation\n");
				return -1;
			}
		}
	}
	if(AdjustPacks)
	{	
		FinalCellAdjust(InputFile,OutputFile);
		UpdateVOBUTable();
	}
	return 0;
}

// This is the beginning of a standard video frame pack.
static uchar FakePack[]={0x0, 0x0, 0x1, 0xba, 0x44, 0x0, 0x4, 0x2e, 0x3d, 0xab, 0x1, 0x89, 0xc3, 0xf8, 
0x0, 0x0, 0x1, 0xe0, 0x7, 0xec, 0x81, 0x0, 0x0};

// This loads a buffer of data from File at lba LBA into buffer.
void LoadBuffer(FILE* File, int LBA, uchar* Buffer)
{
	int ret;
	ret = fseek(File,LBA*SECTOR_SIZE,SEEK_SET); 
	assert(ret == 0);
	ret = fread(Buffer,sizeof(uchar),SECTOR_SIZE,File); 
	assert(ret == SECTOR_SIZE);
}

// This just peeks at the value in File at sector LBA and offset Offset.
uchar PeekValue(FILE* File, int LBA, int Offset)
{
	int ret;
	ret = fseek(File,LBA*SECTOR_SIZE+Offset,SEEK_SET); 
	assert(ret == 0);
	uchar Value;
	ret = fread(&Value,sizeof(uchar),1,File); 
	assert(ret == 1);
	return Value;
}

int CCell::FinalCellAdjust(FILE* InputFile, FILE* OutputFile)
{
	// This is where I do all the final adjustments that require copying from pack to pack.

	// Adjust the PTS of our first I-frame pack, by copying it from the first I-frame after the nav
	// pack we chose as our starting navpack. JEAN: I should check that both PTS and DTS are flagged
	// as present (they should!).
	uchar PTS[10];
	int ret = fseek(InputFile,(m_Packs[0].m_InputLBA+1)*SECTOR_SIZE+0x17,SEEK_SET); assert(ret == 0);
	ret = fread(PTS,sizeof(uchar),10,InputFile); assert(ret == 10);
	fflush(OutputFile);
	ret = fseek(OutputFile,(m_StartLBA+1)*SECTOR_SIZE+0x17,SEEK_SET); assert(ret == 0);
	ret = fwrite(PTS,sizeof(uchar),10,OutputFile); assert(ret == 10);

	// It's important to get the PTS from the I-frame, because the button highlight time will need
	// to match that.
	uint IFramePTS = ParsePTS(PTS);

#if 1
	// Also adjust the PTS of the subpic packs!
	for(int i=0x20;i<0x40;i++)
	{
		int FirstSubpicPack = FindFirstSubPicPack(0,i);
		int j,k;
		while(1)
		{
			// Adjust PTS of all subsequent packs, if they have PTS values (it CAN happen!!!).
			if(FirstSubpicPack == -1)
				break;
			// Check whether PTS data is present.
			if(PeekValue(OutputFile,m_StartLBA+FirstSubpicPack,0x15) & 0x80)
			{
				LoadBuffer(OutputFile,m_StartLBA+FirstSubpicPack,m_Buf);
				PTS[0] &= 0x2f;
				for(j=0x17,k=0;j<=0x1b;j++,k++)
					m_Buf[j] = PTS[k];
				fflush(OutputFile);
				ret = fseek(OutputFile,(m_StartLBA+FirstSubpicPack)*SECTOR_SIZE,SEEK_SET); assert(ret == 0);
				ret = fwrite(m_Buf,sizeof(uchar),SECTOR_SIZE,OutputFile); assert(ret == SECTOR_SIZE);
			}
			FirstSubpicPack = FindNextSubPicPack(FirstSubpicPack,i);
		}
	}
#endif

	// Now, check whether our last I-frame pack was flagged as too full. This would mean that it's a
	// fake video pack which we now need to fill with appropriate data.
	int LastIPack = FindNextEndIFrame(0);
	if(m_Packs[LastIPack].m_IFrameTooFull)
	{
		int i;
		// Load output pack, copy fake pack, but leave eveything up to the pack stuffing length.
		ret = fseek(OutputFile,(m_StartLBA+LastIPack)*SECTOR_SIZE,SEEK_SET); assert(ret == 0);
		ret = fread(m_Buf,sizeof(uchar),SECTOR_SIZE,OutputFile); assert(ret == SECTOR_SIZE);
		for(i=0x0d;i<0x17;i++)
			m_Buf[i] = FakePack[i];
		// now, put a fake picture start code. 
		*((uint*)(m_Buf+i)) = PICT_START_CODE;
		// and call the TrimIFrame function.
		m_Packs[LastIPack].TrimIFrame(m_Buf);
		// Now write pack back!
		ret = fseek(OutputFile,(m_StartLBA+LastIPack)*SECTOR_SIZE,SEEK_SET); assert(ret == 0);
		ret = fwrite(m_Buf,sizeof(uchar),SECTOR_SIZE,OutputFile); assert(ret == SECTOR_SIZE);
	}

	if(m_FixSubPicStops)
	{
		// Experimental code: get rid of subpic stop in subpic packs... These cause all sorts of problems
		// in non-audio mode.
		// Go back to first subpick pack. Then determine length and offset to commands, then fix them
		// This is very ugly because commands can span 2 packs. Eek.
		uchar i=0;
		int j;
		// Try all subpic streams. JEAN: Only use those present.
		for(i=0x20;i<0x40;i++)
		{
			ushort SPLength, SPTrueOffset, SPOffset;
			uchar* Buff2 = NULL;
			// Go to first subpic pack.
			int FirstSubpicPack = FindFirstSubPicPack(0,i);
			if(FirstSubpicPack == -1)
				continue;
			// Load buffer, and find offset to commands.
			LoadBuffer(OutputFile,m_StartLBA+FirstSubpicPack,m_Buf);
			m_Packs[FirstSubpicPack].FindSubpicLengthAndOffset(m_Buf,SPLength,SPOffset,SPTrueOffset);
			// Now, go through subpic packs, subtracting the payload size from SPLength and SPOffset
			j=FirstSubpicPack;
			while(SPOffset >= SECTOR_SIZE && j != -1)
			{
				SPOffset -= (SECTOR_SIZE - 0x18);
				SPLength -= (SECTOR_SIZE - 0x18);
				j=FindNextSubPicPack(j,i);
				if(j!=-1)
				{
					// It sometimes happens that the header length in the following subpacks isn't 0.
					// If it isn't we need to add it to the offset.
					uchar Len = PeekValue(OutputFile,m_StartLBA+j,0x16);
					SPOffset += Len;
					SPLength += Len;
				}
				else
				{
					// This would be very strange. Maybe we're missing the last subpic pack???
					ERROR2("Some data seems to be missing from subpic stream 0x%x\n",i);
					ERRORF2("Il manque des donnes dans le flux de subpic 0x%x\n",i);
					goto end;
				}
			}
			// Load the pack that has the start of the commands area.
			LoadBuffer(OutputFile,m_StartLBA+j,m_Buf);
			// Let's see if the end of the subpic data is in the next subpic pack.
			int j2;
			if(SPLength > SECTOR_SIZE)
			{
				// The commands span 2 subpic packs. let's go to the next subpic pack.
				j2=FindNextSubPicPack(j,i); assert(j2!=-1);
				if(j2==-1)
					goto end;
				// Load that last buffer too.
				Buff2 = new uchar[SECTOR_SIZE];
				LoadBuffer(OutputFile,m_StartLBA+j2,Buff2);
			}
			// This is the function that removes the stop command from the subpic data.
			m_Packs[j].RemoveSubpicStop(m_Buf, Buff2, SPTrueOffset, SPOffset);
			// Write subpic packs back.
			ret = fseek(OutputFile,(m_StartLBA+j)*SECTOR_SIZE,SEEK_SET); assert(ret == 0);
			ret = fwrite(m_Buf,sizeof(uchar),SECTOR_SIZE,OutputFile);
			if(Buff2 != NULL)
			{
				// Write second pack if needed.
				ret = fseek(OutputFile,(m_StartLBA+j2)*SECTOR_SIZE,SEEK_SET); assert(ret == 0);
				ret = fwrite(Buff2,sizeof(uchar),SECTOR_SIZE,OutputFile);
			}
			 assert(ret == SECTOR_SIZE);
end:
			delete[] Buff2;
			fflush(OutputFile);
		}
	}

	// The highlight start time needs to match the PTS of the i-frame, and this in every
	// navpack! GEE! Let's copy button info back to the navpacks in case they miss that.
	if(m_LBAOfFistButtonNavPack != -1)
	{
//		MESSAGE1("Copying button info back to first navpacks\n");
//		MESSAGEF1("Paramtres des boutons copis sur les premiers NavPacks\n");
		// The first navpack does not have button info.
		char ButtonInfo[0x343-0x8d];
		// Load button info from navpack with buttons in the original.
		LoadBuffer(InputFile,m_LBAOfFistButtonNavPack,m_Buf);

		// Don't forget to correct these values. IMPORTANT: Apparently media player classic needs
		// the start highlight time to match exactly the I-frame PTS... Geeee!
		PutInt(m_Buf,0x8f,IFramePTS);
		PutInt(m_Buf,0x93,0xffffffff);
		PutInt(m_Buf,0x97,0xffffffff);
		memcpy(ButtonInfo,m_Buf+0x8d,0x343-0x8d);
		// The first navpack with buttons SHOULD have a status equal to 1.
		assert((ButtonInfo[1] & 0x03) == 1);
		ButtonInfo[1] = 1;
		// Now copy that back in all the preceding navpacks.
		for(i=0;i<m_NumPacks;i++)
		{
			if(m_Packs[i].m_PackType != PackTypeNavPack)
				continue;
			// Write button info.
			if(i!=0)
			{
				// The first button info should have a 1 for the status, all subsequent ones should have 2.
				ButtonInfo[1] = 2;
			}
			int ret = fseek(OutputFile,(m_StartLBA+i)*SECTOR_SIZE+0x8d,SEEK_SET); assert(ret == 0);
			ret = fwrite(ButtonInfo,sizeof(uchar),0x343-0x8d,OutputFile);
			fflush(OutputFile);
		}
	}

#if 0
	// Adjust cell elapsed times. JEAN: THIS ISN"T NEEDED ANYMORE!
	CDVDTime InitTime;
	CDVDTime NavPackTime;
	// Read elapsed time in first NavPack.
	LoadBuffer(OutputFile,m_StartLBA,m_Buf);
	InitTime.ReadFromBuffer(m_Buf+0x45);
	if(!InitTime.IsZero())
	{
		// The first navpack has a non zero elapsed time. Let's subtract it from all subsequent ones.
		MESSAGE1("Updating cell elapsed times\n");
		MESSAGEF1("Modification de la dure de la cellule\n");
		int i;
		for(i=0;i<m_NumPacks;i++)
		{
			if(m_Packs[i].m_PackType != PackTypeNavPack)
				continue;
			// A bit of a waste to read a whole buffer. JEAN: FIX
			LoadBuffer(OutputFile,m_StartLBA+i,m_Buf);
			NavPackTime.ReadFromBuffer(m_Buf+0x45);
			NavPackTime -= InitTime;
			NavPackTime.WriteToBuffer(m_Buf+0x45);
			NavPackTime.WriteToBuffer(m_Buf+0x423);
			int ret = fseek(OutputFile,(m_StartLBA+i)*SECTOR_SIZE+0x45,SEEK_SET); assert(ret == 0);
			ret = fwrite(m_Buf+0x45,sizeof(uchar),4,OutputFile); 
		}
	}
#endif

	ret = fseek(OutputFile,0,SEEK_END); assert(ret == 0);
	return 0;
}

int CCell::InsertPack(CPack* ThisPack, int Index)
{
	// Move all the packs after this one up, then put this one in!
	if(m_NumPacks + 1 > m_NumAllocPacks)
	{
		// ERROR!
		return -1;
	}

	int i;
	// Move all the packs downstream.
	for(i=m_NumPacks;i>Index;i--)
	{
		m_Packs[i] = m_Packs[i-1];
	}
	m_Packs[Index] = *ThisPack;

	if(m_Packs[Index].m_PackType == PackTypeNavPack)
	{
		m_NavPackCount++;
		if(m_FirstButtonIndex == -1 && m_Packs[Index].m_NumButtons > 0)
		{
			m_FirstButtonIndex = Index;
		}
	}
	if(m_Packs[Index].m_PackType == PackTypeAudio)
	{
		// Keep track of which audio streams are present. 
		int St;
		for(St=0;St<m_NumFoundAudioStreams;St++)
		{
			if(m_Packs[Index].m_StreamID == m_FoundAudioStreams[St])
				break;
		}
		if(St == m_NumFoundAudioStreams)
		{
			assert(St < 8);
			m_FoundAudioStreams[m_NumFoundAudioStreams++] = m_Packs[Index].m_StreamID;
		}
	}
	m_NumPacks++;
	return 0;
}

int CCell::AppendPack(CPack* ThisPack)
{
	return InsertPack(ThisPack, m_NumPacks);
}


int CCell::LoadPack(FILE* InputFile, int PackIndex)
{
	int ret;
	if(m_InputImage != 0 && m_Packs[PackIndex].m_PackType == PackTypeVideo)
		InputFile = m_InputImage;
	int InLBACount = m_Packs[PackIndex].m_InputLBA;
	ret = fseek(InputFile,InLBACount*SECTOR_SIZE,SEEK_SET);
	if(ret != 0)
	{
		// ERROR!
		ERROR2("Couldn't seek to sector %d, exiting \n",InLBACount);
		ERRORF2("Impossible d'aller au secteur %d, annulation\n",InLBACount);
		return -1;
	}
	// Read one LBA.
	ret = fread(m_Buf,sizeof(uchar),SECTOR_SIZE,InputFile);
	if(ret != SECTOR_SIZE)
	{
		ERROR2("Couldn't read sector %d, exiting \n",InLBACount);
		ERRORF2("Impossible de lire le secteur %d, annulation\n",InLBACount);
		return -1;
	}
	return 0;
}

int CCell::AdjustAndWritePack(FILE* OutputFile, int PackIndex)
{
	int OutLBACount = m_StartLBA + PackIndex;
	int ret,i;

	// STILL TO DO:
	// IN SML_PBI (seamless playback). Adjust 0433 to PTM of first VOBU. THis shouldn't change though.
	// PTM of last video frame in last GOP of VOB. PROBLEM! Can't do that on a cell/cell basis!
	// SYNCI Offsets to Audio Packs! Dont understand how it could be negative. Do streams 2 and up.
	// SYNCI to subpics. A bit of a mess. What if the subpics are NOT in the first VOBU, then what?

	// Now let's adjust what needs to be adjusted.
	CPack* ThisPack = &m_Packs[PackIndex];
	if(ThisPack->m_PackType == PackTypeNavPack)
	{
//		if(PackIndex)
//			MESSAGE2("Found next NavPack at LBA %d\n",ThisPack->m_InputLBA);
		// Adjust nav pack!
		// Adjust the this->lba at 0x040b and 0x2d. Should be OutLBACount.
		PutInt(m_Buf,0x2d,OutLBACount);
		PutInt(m_Buf,0x040b,OutLBACount);
		// Adjust the VOBU end address at 0x040f. It's relative to the end of this LBA.
		int NextNavPackLBA = FindNextNavPack(PackIndex);
		if(NextNavPackLBA == -1)
			NextNavPackLBA = m_NumPacks;
		NextNavPackLBA += m_StartLBA; 
		PutInt(m_Buf,0x040f,NextNavPackLBA - OutLBACount -1);

		// Now for the first second and third references frame end blocks
		// Set the second and third references frame end blocks to 0.
		int FirstRef = 0;
		if(PackIndex == 0)
		{
			// For the first pack, this should point to the end of the Iframe.
			FirstRef = FindNextEndIFrame(PackIndex) - PackIndex;
			if(FirstRef < 0)
			{
				MESSAGE2("NOTE: Can't find end of I-frame at input LBA %d\n",ThisPack->m_InputLBA);
				MESSAGEF2("NOTE: Impossible de trouver la fin de la I-frame au secteur LBA %d\n",ThisPack->m_InputLBA);
				FirstRef = 0;
			}
		}
		PutInt(m_Buf,0x0413,FirstRef);
		PutInt(m_Buf,0x0417,0);
		PutInt(m_Buf,0x041b,0);

		if(PackIndex == 0)
		{
			// Fix the end presentation time for the end_of_sequence. That's because this nav pack
			// is the one that corresponds to the I-frame!!!!
			uint StartPTS = GetInt(m_Buf,0x39);
			m_PTSFirstNavPack = StartPTS;
			assert(VideoFormat != FORMAT_UNKNOWN);
			if(VideoFormat == FORMAT_UNKNOWN)
			{
				MESSAGE1("WARNING: Unknown video format while adjusting End of Presentation Time\n");
				MESSAGEF1("ATTENTION: Le format video n'a pas t trouv\n");
			}
			PutInt(m_Buf,0x41,StartPTS + ((VideoFormat == FORMAT_NTSC) ? 3003 : 3600));
			// If there's only 1 navpack, I should really adjust the end PTS of the VOBU.
			if(FindNextNavPack(PackIndex) == -1 && m_FoundButtons == 0)
			{
				PutInt(m_Buf,0x3d,StartPTS + ((VideoFormat == FORMAT_NTSC) ? 3003 : 3600));
			}
		}

		// Set the offset to the next VOBU with video to bfffffff
		PutInt(m_Buf,0x04f1,0xbfffffff);

		// If not in the first NavPack, set offset of the previous VOBU with video.
		if(PackIndex == 0)
			PutInt(m_Buf,0x0595,0xbfffffff);
		else
			PutInt(m_Buf,0x0595,0x80000000+PackIndex);
		// JEAN! Copying JSOTO. Previous vobu with video.
		PutInt(m_Buf,0x0595,0xbfffffff);
		// Let's set the PTM of last video in last gop (slplay 0x437) to match end ptm of VOBU
		if(!m_KeepAudio)
		{
			int ENDPTMofVOBU = GetInt(m_Buf,0x3d);
			PutInt(m_Buf,0x437,ENDPTMofVOBU);
			int StartPTMofVOBU = GetInt(m_Buf,0x39);
			PutInt(m_Buf,0x433,StartPTMofVOBU);
		}

		if(!m_KeepAudio)
		{
			// VOBU_SRI: Next VOBUS with video.
			// If there's no audio, none of these other pointers should be set.
			for(i=0x04f5;i<=0x591;i+=4)
			{
				PutInt(m_Buf,i,0x3fffffff);
			}
		}
		else
		{
			// If there's audio.
			for(i=0x04f5;i<=0x591;i+=4)
			{
				PutInt(m_Buf,i,0x3fffffff);
			}
			// Next VOBU with possible video. This seems important for some DVD recorders...
			// BUT I'm finding that this can't point to the next cell. Mediaplayerclassic
			// hangs otherwise...
			NextNavPackLBA = FindNextNavPack(PackIndex);
			if(NextNavPackLBA != -1)
			{
				NextNavPackLBA += m_StartLBA; 
				PutInt(m_Buf,0x53d,0x80000000+NextNavPackLBA - OutLBACount);
				PutInt(m_Buf,0x541,0x80000000+NextNavPackLBA - OutLBACount);
			}
			if(PackIndex != 0)
			{
				// Previous VOBU with possible video.
				int PrevNavPack = FindPrevNavPack(PackIndex);
				assert(PrevNavPack != -1);
				PutInt(m_Buf,0x545,0x80000000+PackIndex - PrevNavPack);
				PutInt(m_Buf,0x549,0x80000000+PackIndex - PrevNavPack);
			}
		}

		// SYNCI Stuff!
		// Invalidate all the SYNCI audio pointers. These are SHORTS!
		for(i=0x0599;i<=0x5a7;i+=2)
		{
			PutShort(m_Buf,i,0x0000);
		}
		// ONLY DOING THE FIRST STREAM. FIX! JEAN
		int NextAudioPack = FindNextAudioPackInVOBU(PackIndex);
		if(NextAudioPack != -1)
			PutShort(m_Buf,0x0599,NextAudioPack-PackIndex);
		// SUBPIC offsets: Points to navpack of first VOBU with subpic. In this case, it would be
		// simply pointing to the first NavPack, but only if there are subpics.
//PutInt(m_Buf,0x05a9,0x80000000+PackIndex);


		// Button highlight times! FIX THIS!
		if(ThisPack->m_NumButtons) // Philips says it must be -1 for a still.
		{
			assert(m_PTSFirstNavPack != 0);
			PutInt(m_Buf,0x8f,m_PTSFirstNavPack);
			PutInt(m_Buf,0x93,0xffffffff);
			PutInt(m_Buf,0x97,0xffffffff);
		}
	}

	if(ThisPack->m_IFrameType == IFrameTypeStart)
	{
		// Adjust GOP flags.
		ThisPack->CloseGOP(m_Buf);
	}

	// For the last I-frame, and if it's not too full, let's remove stuff we no longer need, and add
	// an end_of_sequence. If it's too full, we'll do that at FinalCellAdjust() time.
	if(ThisPack->m_IFrameType == IFrameTypeEnd && !ThisPack->m_IFrameTooFull)
	{
		// Adjust last I-frame packet.
		ThisPack->TrimIFrame(m_Buf);
	}

	if(m_AdjustSRC && ThisPack->m_FrameType == FrameTypeI)
	{
		// We need to write the SRC of the I-frame packs. Let's do it.
		PutShort(m_Buf,0x4,ThisPack->m_SRCH);
		PutInt(m_Buf,0x6,ThisPack->m_SRCL);
	}

	if(ThisPack->m_PackType == PackTypeSupPic)
	{		
//		MESSAGE3("Keeping subpic pack 0x%x at LBA %d\n",ThisPack->m_StreamID,ThisPack->m_InputLBA);
		// Let's write the SRC back. 
		PutShort(m_Buf,0x4,ThisPack->m_SRCH);
		PutInt(m_Buf,0x6,ThisPack->m_SRCL);
	}

	// Write 1 LBA:
	ret = fwrite(m_Buf,sizeof(uchar),SECTOR_SIZE,OutputFile);
	if(ret != SECTOR_SIZE)
	{
		ERROR2("Couldn't write sector %d, exiting \n",OutLBACount);
		ERRORF2("Impossible d'crire le secteur %d, annulation\n",OutLBACount);
		return -1;
	}

	return 0;
}

void CCell::UpdateVOBUTable()
{
	// First, allocate the vobu table.
	delete[] m_VOBUTable;
	m_VOBUTable = new int[m_NavPackCount];
	int i,j;
	for(i=j=0;i<m_NumPacks;i++)
	{
		if(m_Packs[i].m_PackType == PackTypeNavPack)
		{
			if(j == m_NavPackCount)
			{
				ERROR1("Error while updating VOBU Table!\n");
				ERRORF1("Erreur pendant la modification de la table des VOBUs!\n");
				return;
			}
			m_VOBUTable[j++] = m_StartLBA+i;
		}
	}
}

int CCell::FindNextNavPack(int CurrentPackIndex)
{
	int i;
	for(i=CurrentPackIndex+1;i<m_NumPacks;i++)
	{
		if(m_Packs[i].m_PackType == PackTypeNavPack)
			return i;
	}
	return -1;
}

int CCell::FindPrevNavPack(int CurrentPackIndex)
{
	int i;
	for(i=CurrentPackIndex-1;i>=0;i--)
	{
		if(m_Packs[i].m_PackType == PackTypeNavPack)
			return i;
	}
	return -1;
}

int CCell::FindNextAudioPackInVOBU(int CurrentPackIndex)
{
	int i;
	for(i=CurrentPackIndex+1;i<m_NumPacks;i++)
	{
		// In the same VOBU, so look until next nav pack.
		if(m_Packs[i].m_PackType == PackTypeNavPack)
			return -1;
		if(m_Packs[i].m_PackType == PackTypeAudio)
			return i;
	}
	return -1;
}

int CCell::FindNextSubPicPack(int CurrentPackIndex, uchar StreamID)
{
	int i;
	if(StreamID == 0)
	{
		for(i=CurrentPackIndex+1;i<m_NumPacks;i++)
		{
			if(m_Packs[i].m_PackType == PackTypeSupPic)
				return i;
		}
		return -1;
	}
	else
	{
		for(i=CurrentPackIndex+1;i<m_NumPacks;i++)
		{
			if(m_Packs[i].m_PackType == PackTypeSupPic && m_Packs[i].m_StreamID == StreamID)
				return i;
		}
		return -1;
	}
}

int CCell::FindLastSubPicPack(int CurrentPackIndex, uchar StreamID)
{
	int i;
	if(StreamID == 0)
	{
		for(i=m_NumPacks-1;i>CurrentPackIndex;i--)
		{
			if(m_Packs[i].m_PackType == PackTypeSupPic)
				return i;
		}
		return -1;
	}
	else
	{
		for(i=m_NumPacks-1;i>CurrentPackIndex;i--)
		{
			if(m_Packs[i].m_PackType == PackTypeSupPic && m_Packs[i].m_StreamID == StreamID)
				return i;
		}
		return -1;
	}
}

int CCell::FindFirstSubPicPack(int CurrentPackIndex, uchar StreamID)
{
	int i;
	if(StreamID == 0)
	{
		for(i=1;i<m_NumPacks;i++)
		{
			if(m_Packs[i].m_PackType == PackTypeSupPic)
				return i;
		}
		return -1;
	}
	else
	{
		for(i=1;i<m_NumPacks;i++)
		{
			if(m_Packs[i].m_PackType == PackTypeSupPic && m_Packs[i].m_StreamID == StreamID)
				return i;
		}
		return -1;
	}
}

int CCell::CountIFrames(int StartIndex, int EndIndex)
{
	int i;
	int IFrameCount = 0;
	for(i=StartIndex;i<EndIndex;i++)
	{
		if(m_Packs[i].m_IFrameType == IFrameTypeStart && m_Packs[i].m_IFrameSize > 0)
			IFrameCount++;
	}
	return IFrameCount;
}

int CCell::FindThisIFrame(int IFrame, int StartLBA)
{
	int i;
	int IFrameCount = 0;
	if(IFrame >= 0)
	{
		for(i=StartLBA;i<m_NumPacks;i++)
		{
			if(m_Packs[i].m_IFrameType == IFrameTypeStart && m_Packs[i].m_IFrameSize > 0)
			{
				if(IFrameCount == IFrame)
					return i;
				IFrameCount++;
			}
		}
		return -1;
	}
	else
	{
		IFrame = -IFrame -1;
		for(i=m_NumPacks-1;i>=StartLBA;i--)
		{
			if(m_Packs[i].m_IFrameType == IFrameTypeStart && m_Packs[i].m_IFrameSize > 0)
			{
				if(IFrameCount == IFrame)
					return i;
				IFrameCount++;
			}
		}
		return -1;
	}
}

int CCell::FindNextIFrame(int StartLBA)
{
	int i;
	for(i=StartLBA+1;i<m_NumPacks;i++)
	{
		if(m_Packs[i].m_IFrameType == IFrameTypeStart && m_Packs[i].m_IFrameSize > 0)
			return i;
	}
	return -1;
}

// Find the largest I-Frame.
int CCell::FindLargestIFrame(int StartLBA)
{
	short MaxSize = 0;
	int MaxIndex = -1;
	int i;
	for(i=StartLBA;i<m_NumPacks;i++)
	{
		if(m_Packs[i].m_IFrameType == IFrameTypeStart && m_Packs[i].m_IFrameSize > MaxSize)
		{
			MaxSize = m_Packs[i].m_IFrameSize;
			MaxIndex = i;
		}
	}
	assert(MaxIndex != -1);
	return MaxIndex;
}

int CCell::FindSmallestIFrame(int StartLBA)
{
	short MinSize = 32000;
	int MinIndex = -1;
	int i;
	for(i=StartLBA;i<m_NumPacks;i++)
	{
		if(m_Packs[i].m_IFrameType == IFrameTypeStart && m_Packs[i].m_IFrameSize < MinSize)
		{
			MinSize = m_Packs[i].m_IFrameSize;
			MinIndex = i;
		}
	}
	assert(MinIndex != -1);
	assert(MinSize > 0);
	return MinIndex;
}

int CCell::FindNextEndIFrame(int CurrentPackIndex)
{
	int i;
	for(i=CurrentPackIndex+1;i<m_NumPacks;i++)
	{
		if(m_Packs[i].m_IFrameType == IFrameTypeEnd)
			return i;
	}
	if(m_Packs[i-1].m_IFrameType == IFrameTypeStart)
	{
		// special case, the start IS also the end! That can happen!
		return i-1;
	}
	return -1;
}

void CCell::ResetCell()
{
	memset(m_Packs, 0, m_NumAllocPacks*sizeof(CPack));
	m_StartLBA = 0;
	m_EndLBA = 0;
	m_VOBID = 0;		
	m_CELLID = 0;		
	m_StartPTS = 0;		
	m_EndPTS = 0;		
	m_NavPackCount = 0;	
	m_KeepAudio = 0;
	m_FoundButtons = 0;
	m_IFrameCount = 0;
	m_NumPacks = 0;	
}

int CCell::CopyFirstIFrameNoAudio(CCell* SrcCell)
{
	// First of all, blank out the cell.
	ResetCell();
	m_KeepAudio = 0;
	m_AdjustSRC = 0;
	m_VOBID = SrcCell->m_VOBID;
	m_CELLID = SrcCell->m_CELLID;
	m_FoundButtons = SrcCell->m_FoundButtons;
	
	int InIndex = 0;

	// Copy first NavPack.
	AppendPack(&SrcCell->m_Packs[0]);
	for(InIndex=1;InIndex<SrcCell->m_NumPacks;InIndex++)
	{
		// Stop at the next NavPack!
		if(SrcCell->m_Packs[InIndex].m_PackType == PackTypeNavPack)
			break;
		// Don't copy Audio packs.
		if(SrcCell->m_Packs[InIndex].m_PackType == PackTypeAudio)
			continue;
		// Don't copy video packs other than I-frame.
		if(SrcCell->m_Packs[InIndex].m_FrameType != FrameTypeI && 
			SrcCell->m_Packs[InIndex].m_PackType != PackTypeSupPic)
			continue;
		// Copying I-frames, subpic frames...
		AppendPack(&SrcCell->m_Packs[InIndex]);
	}

#if 0
	// This is not going to create a legitimate cell because there will be large gaps between VOBUs.
	int NextSubPickPack ;
	while((NextSubPickPack = SrcCell->FindNextSubPicPack(InIndex)) != -1)
	{
		// There's a subpic pack somewhere... Let's copy all the navpacks until we find it.
		for(;InIndex<=NextSubPickPack;InIndex++)
		{
			// Don't copy Audio or video packs.
			if(SrcCell->m_Packs[InIndex].m_PackType == PackTypeAudio || 
				SrcCell->m_Packs[InIndex].m_PackType == PackTypeVideo)
				continue;
			AppendPack(&SrcCell->m_Packs[InIndex]);
		}
	}
#endif

#if 0
	// Another option, keep all the navpacks...
	for(;InIndex<SrcCell->m_NumPacks;InIndex++)
	{
		// Don't copy Audio or video packs.
		if(SrcCell->m_Packs[InIndex].m_PackType == PackTypeAudio || 
			SrcCell->m_Packs[InIndex].m_PackType == PackTypeVideo)
			continue;
		AppendPack(&SrcCell->m_Packs[InIndex]);
	}
#else
#endif
	// We're done.
	return 0;
}


int CCell::CopyFirstIFrameWithAudio(CCell* SrcCell)
{
	// First of all, blank out the cell.
	ResetCell();
	m_KeepAudio = 1;
	m_AdjustSRC = 0;
	m_VOBID = SrcCell->m_VOBID;
	m_CELLID = SrcCell->m_CELLID;
	m_FoundButtons = SrcCell->m_FoundButtons;
	
	int InIndex = 0;

	// Copy first NavPack.
	AppendPack(&SrcCell->m_Packs[0]);
	for(InIndex=1;InIndex<SrcCell->m_NumPacks;InIndex++)
	{
		// Stop at the next NavPack!
		if(SrcCell->m_Packs[InIndex].m_PackType == PackTypeNavPack)
			break;
		// Don't copy video packs other than I-frame.
		if(SrcCell->m_Packs[InIndex].m_PackType == PackTypeVideo &&
			SrcCell->m_Packs[InIndex].m_FrameType != FrameTypeI)
			continue;
		// Copying I-frames, subpic frames and audio packs...
		AppendPack(&SrcCell->m_Packs[InIndex]);
	}
	// Now go on with the rest of the cell, but skip any video.
	for(;InIndex<SrcCell->m_NumPacks;InIndex++)
	{
		// Don't copy Audio or video packs.
		if(SrcCell->m_Packs[InIndex].m_PackType == PackTypeVideo)
			continue;
		AppendPack(&SrcCell->m_Packs[InIndex]);
	}
	// We're done.
	return 0;	
}

int CCell::CopySingleIFrame(CCell* SrcCell, int SrcIFrame, bool KeepAudio)
{
	// First of all, blank out the cell.
	ResetCell();
	m_KeepAudio = KeepAudio;
	m_AdjustSRC = 1;
	int InIndex = 0;
	int IFrameIndex = 0;
	int i;
	int NumIFrames = 0;

	CCell* ImageCell = SrcCell;

	m_VOBID = SrcCell->m_VOBID;
	m_CELLID = SrcCell->m_CELLID;
	m_StartPTS = SrcCell->m_StartPTS;
	m_EndPTS = SrcCell->m_EndPTS;
	m_FoundButtons = SrcCell->m_FoundButtons;
	m_MaxSizeX = SrcCell->m_MaxSizeX;
	m_MaxSizeY = SrcCell->m_MaxSizeY;
	m_AllNops = SrcCell->m_AllNops;
	m_ForcedAction = SrcCell->m_ForcedAction;
	m_AnyLinkNext = SrcCell->m_AnyLinkNext;
	m_PreviewAllCells = SrcCell->m_PreviewAllCells;
	memcpy(m_NumberSubpicUnits,SrcCell->m_NumberSubpicUnits,32*sizeof(int));

	int StartNavPack = 0;
	if(m_FoundButtons == 0)
	{
		m_KeepAudio = KeepAudio = 0;
	}
	else if(m_MaxSizeX < m_MinButtonSize || m_MaxSizeY < m_MinButtonSize)
	{
		if(m_ForcedAction == 0)
		{
			MESSAGE3("NOTE: buttons are tiny (%dx%d) killing the cell.\n",m_MaxSizeX,m_MaxSizeY);
			MESSAGEF3("Les boutons sone minuscules (%dx%d). La cellule va etre traite comme une intro.\n",m_MaxSizeX,m_MaxSizeY);
			m_FoundButtons = KeepAudio = m_KeepAudio = 0;
		}
		else
		{
			MESSAGE3("WARNING: THIS CELL HAS TINY BUTTONS! Max Size %dx%d\n",m_MaxSizeX,m_MaxSizeY);
			MESSAGEF3("ATTENTION! LES BOUTONS SONT MINUSCULES: Taille max: %dx%d\n",m_MaxSizeX,m_MaxSizeY);
		}

	}
	else if(m_AllNops)
	{
		MESSAGE1("NOTE: All buttons in this cell have NOP commands, killing the cell\n");
		MESSAGEF1("NOTE: Tous les boutons ont NOP pour commande. La cellule va etre traite comme une intro.\n");
		m_FoundButtons = KeepAudio = m_KeepAudio = 0;
	}

	// More warnings if we find more than 1 SPU (Subpicture unit) per stream.
	bool Any = 0;
	for(i=0;i<32;i++)
	{
		if(m_NumberSubpicUnits[i] > 1)
		{
			// I'm assuming if there's more than 4 SPUs, we have animated menus...
			if(m_NumberSubpicUnits[i] > 4)
				Any = 1;
			MESSAGE3("NOTE: found %d subpicture units in stream 0x%x!\n", m_NumberSubpicUnits[i],0x20+i);
			MESSAGEF3("NOTE: il y a %d units de subpic dans le flux 0x%x!\n",m_NumberSubpicUnits[i],0x20+i);
		}
	}
	if(Any)
	{
		MESSAGE1("NOTE: This cell probably has animated highlights.\nThey will not be animated in the processed DVD.\n");
		MESSAGEF1("NOTE: Ce menu a des subpics anims. Ils seront statiques\ndans le DVD final.\n");
	}

	// First find the Iframe in question. It's a good idea to find a frame after the subpics appear
	// otherwise the background might not show any buttons...
	int FirstIframe = SrcCell->m_FirstButtonIndex;
	if(FirstIframe == -1)
		FirstIframe = 0;
#if 0
	if(FirstIframe > 0)
	{
		MESSAGE2("NOTE: Buttons appear late, at LBA %d\n",SrcCell->m_StartLBA+FirstIframe);
		MESSAGEF2("NOTE: Les boutons apparaissent au milieu de la cellule, au secteur %d\n",SrcCell->m_StartLBA+FirstIframe);
	}
#endif

	if(SrcIFrame == 9999)		// JEAN: UGLY, FIX!
	{
		// Special case, let's find the largest one.
		IFrameIndex = SrcCell->FindLargestIFrame(FirstIframe);
	}
	else
	{
		// Find the requested I-Frame. Rather than taking the very first I-frame, or the very last one,
		// a better idea is to take one a little bit into the cell, or close to but not exactly at the
		// end of the cell. Let's find out how many I-frame we have.
		NumIFrames = SrcCell->CountIFrames(FirstIframe,SrcCell->m_NumPacks);
		if(NumIFrames == 0)
		{
			FirstIframe = 0;
			NumIFrames = SrcCell->CountIFrames(FirstIframe,SrcCell->m_NumPacks);		
		}
		// Select an I-frame 15% into the cell or 85% into it instead.
		assert(NumIFrames > 0);
		SrcIFrame = (int) (((NumIFrames-1) * SrcIFrame) / 100);
		IFrameIndex = SrcCell->FindThisIFrame(SrcIFrame,FirstIframe);
		if(IFrameIndex == -1)
		{
			ERROR2("Couldn't find IFrame %d \n",SrcIFrame);
			ERRORF2("Impossible de trouver l'image %d \n",SrcIFrame);
			assert(0);
			return -1;
		}
		if(m_FoundButtons == 0 && (!m_PreviewAllCells || NumIFrames <= 2 || FrameSelectFunction == 0))
		{
			// Choose the smallest frame if there's no button, and we don't want to preview all the
			// cells. If we do want to preview all the cells, do it still if there's less than 2 I-frames.
			IFrameIndex = SrcCell->FindSmallestIFrame(0);
		}
		if(NumIFrames > 2 && (m_FoundButtons == 1 || m_PreviewAllCells))
		{
			// If there's more than a couple frames, and there are buttons, we'll call the FrameSelectFuction
			// if it's non-null. This is a way for another program to intervene in the choice of the I-frame.
			// For example, you could pop up a dialog with a preview and let the user select.
			int LastIFrame = SrcCell->FindThisIFrame(-1, 0);
			if(FrameSelectFunction)
			{
				IFrameIndex = FrameSelectFunction(m_VOBID,m_CELLID,SrcCell->m_Packs[FirstIframe].m_InputLBA,
					SrcCell->m_Packs[LastIFrame].m_InputLBA, SrcCell->m_Packs[IFrameIndex].m_InputLBA,
					SrcCell->m_NumPacks,!m_FoundButtons);

				if(IFrameIndex >= 0)
				{
					// I need this just in case the frame returned isn't quite accurate.
					IFrameIndex -= SrcCell->m_StartLBA;
					IFrameIndex -= 2; if(IFrameIndex < 0) IFrameIndex = 0;
					IFrameIndex = SrcCell->FindThisIFrame(0,IFrameIndex);
				}
				else if(IFrameIndex == -1)
				{
					// If FrameSelectFunction() returns -1, the user wants to skip this cell.
					m_FoundButtons = 0;
					m_KeepAudio = KeepAudio = 0;
					IFrameIndex = SrcCell->FindSmallestIFrame(0);
				}
				else if(IFrameIndex == -2)
				{
					// If FrameSelectFunction() returns -2, the user wants to keep this cell intact.
					MESSAGE1("Keeping this cell animated.\n");
					MESSAGEF1("L'animation de la cellule est prserve\n");
					return -2;
				}

#if 0
				static int first = 1;
				if(1||first == 1)
				{
					FILE* fd = fopen("E:/Temp/arenitazul.VOB","rb");
					m_InputImage = fd;
					CCell* TheImage = new CCell(m_InputImage);
					int UU;
					TheImage->ReadFromFile(m_InputImage,UU);
					ImageCell = TheImage;
					IFrameIndex = 1;
				}
				first = 0;
#else
#endif
			}
		}
	}

	MESSAGE2("Inserting I-Frame at original LBA %d \n",ImageCell->m_Packs[IFrameIndex].m_InputLBA);
	MESSAGEF2("Insertion de l'image trouve au secteur %d \n",ImageCell->m_Packs[IFrameIndex].m_InputLBA);
	
	// Copy first NavPack and first I-frame video pack.
	AppendPack(&SrcCell->m_Packs[StartNavPack]);
	InIndex=IFrameIndex;
	AppendPack(&ImageCell->m_Packs[InIndex++]);
	// Put all subpic packs we can find in the cell. That's really where they should be for a still.
#if 0
	for(i=0;i<SrcCell->m_NumPacks;i++)
	{
		if(SrcCell->m_Packs[i].m_PackType == PackTypeSupPic)
			AppendPack(&SrcCell->m_Packs[i]);
	}
#else
	// JEAN: I put this to prevent inserting too many subpic packs after the first video pack.
	// I need many more tests to make sure this does not break anything.
	int StopHere[32];
	memset(StopHere,0,32*sizeof(bool));
	for(i=0;i<SrcCell->m_NumPacks;i++)
	{
		if(SrcCell->m_Packs[i].m_PackType == PackTypeSupPic)
		{
			// Count how many SPUs we have, for each stream. Stop after inserting a certain number...
			if(SrcCell->m_Packs[i].m_NumButtons == 1)
				StopHere[SrcCell->m_Packs[i].m_StreamID-0x20]++;
			if(StopHere[SrcCell->m_Packs[i].m_StreamID-0x20] > 4)
				continue;
			AppendPack(&SrcCell->m_Packs[i]);
		}
	}
#endif
	// Then the rest of the I-frame:
	for(;InIndex<ImageCell->m_NumPacks;InIndex++)
	{
		// Stop at the next NavPack!
		if(ImageCell->m_Packs[InIndex].m_PackType == PackTypeNavPack)
			break;
		// Don't copy anything but I-frame packs
		if(ImageCell->m_Packs[InIndex].m_FrameType != FrameTypeI)
			continue;
		AppendPack(&ImageCell->m_Packs[InIndex]);
	}

	// Now check whether this I-frame is usable as is, i.e., we can put our end_of_sequence code at the
	// end of it!
	if(m_Packs[m_NumPacks-1].m_IFrameTooFull)
	{
		// The last pack of the I-Frame is too full to let us put our end_of_sequence code. We need to
		// insert a dummy video frame, which we'll fill up at write time. I'm simply using pack 0 of
		// the original cell, but faking it as a video cell.
		AppendPack(&ImageCell->m_Packs[0]);
		m_Packs[m_NumPacks-1].m_PackType = PackTypeVideo;
		m_Packs[m_NumPacks-1].m_FrameType = FrameTypeI;
		m_Packs[m_NumPacks-1].m_IFrameType = IFrameTypeEnd;
		m_Packs[m_NumPacks-1].m_IFrameTooFull = 1;
		// Let's also change the type of the preceding pack.
		m_Packs[m_NumPacks-2].m_IFrameType = IFrameTypeMiddle;
		InIndex++;
	}

	// OK, the I-frame and the subpics are in there. Now we need to put the audio that was in the
	// original first VOBU, it necessary. 
	CompositeSRC CurrentSRC;
	// Set the current SRC to the Nav pack's and Increment it by the right value.
	SrcToFrac(m_Packs[0].m_SRCL,m_Packs[0].m_SRCH,CurrentSRC);
	IncSRC(CurrentSRC);
	bool DoneWithIFrame = 0;
	bool DoneWithOthers = 0;
	// Set the SRC of the firt I-frame pack, it HAS to be the first one after the nav pack!
	int IFIndex = 2;
	FracToSrc(CurrentSRC,m_Packs[1].m_SRCL,m_Packs[1].m_SRCH);
	IncSRC(CurrentSRC);
	i = StartNavPack+1;
	while(DoneWithIFrame == 0 || DoneWithOthers == 0)
	{
		CompositeSRC InsertPacSrc;
		// Look for audio packs in our first VOBU.
		for(;i<=SrcCell->m_NumPacks;i++)
		{
			// Stop at the next NavPack or at the end of the cell;
			if(i==SrcCell->m_NumPacks || SrcCell->m_Packs[i].m_PackType == PackTypeNavPack)
			{
				DoneWithOthers = 1;
				break;
			}
			// Don't copy anything but audio packs.
			if(!KeepAudio || SrcCell->m_Packs[i].m_PackType != PackTypeAudio)
				continue;
			break;
		}
		if(!DoneWithOthers)
		{
			// We have one audio pack to insert somewhere. Let's get its SRC.
			SrcToFrac(SrcCell->m_Packs[i].m_SRCL,SrcCell->m_Packs[i].m_SRCH,InsertPacSrc);
		}
		for(;IFIndex<m_NumPacks;IFIndex++)
		{
			// We can leave this I-Frame or subpic pack here if the current SRC + inc is still 
			// below that of the audio to insert
			CompositeSRC NextSRC = CurrentSRC;
			IncSRC(NextSRC);
			if(DoneWithOthers || NextSRC.m_SRCInt < InsertPacSrc.m_SRCInt || 
				(NextSRC.m_SRCInt == InsertPacSrc.m_SRCInt && NextSRC.m_SRCFrac <= InsertPacSrc.m_SRCFrac))
			{
				// We're good. Simply set the SRC of the I-Frame/subpic pack.
				FracToSrc(CurrentSRC,m_Packs[IFIndex].m_SRCL,m_Packs[IFIndex].m_SRCH);
				IncSRC(CurrentSRC);
			}
			else
			{
				// We must insert the audio pack here!
				break;
			}
		}
		// Let's see if we're done setting the SRC of all the I-frame and subpic packs.
		if(IFIndex == m_NumPacks)
			DoneWithIFrame = true;
		if(!DoneWithOthers)
		{
			// We must put the audio in there.
			InsertPack(&SrcCell->m_Packs[i],IFIndex);
			// increment IFIndex so it points to the I-frame/subpic pack right after this one...
			IFIndex++;
			// Set current SRC to match the audio pack SRC and increment.
			CurrentSRC = InsertPacSrc;
			IncSRC(CurrentSRC);
			i++;
		}
	}

	if(SrcCell->m_FirstButtonIndex != -1)
	{
		if(SrcCell->m_LastButtonChange != SrcCell->m_FirstButtonIndex)
		{
			MESSAGE2("NOTE: Button definition changed mid-cell at LBA %d\n",SrcCell->m_Packs[SrcCell->m_LastButtonChange].m_InputLBA);
			MESSAGEF2("NOTE: Les boutons changent au milieu de la cellule au secteur %d\n",SrcCell->m_Packs[SrcCell->m_LastButtonChange].m_InputLBA);
		}
		// I need to remember that LBA so I can copy the button information back. That's because the
		// button info might appear much later in the cell than in the first VOBU.
		m_LBAOfFistButtonNavPack = SrcCell->m_Packs[SrcCell->m_LastButtonChange].m_InputLBA;
	}

	if(KeepAudio)
	{
		// Now go on with the rest of the cell, but only keep audio packs and navpacks.
		for(;i<SrcCell->m_NumPacks;i++)
		{
			// Don't copy video packs.
			if(SrcCell->m_Packs[i].m_PackType == PackTypeVideo || SrcCell->m_Packs[i].m_PackType == PackTypeSupPic)
				continue;
			// Here, I'm including the subpic packs, in case there are animated highlights.
//			if(SrcCell->m_Packs[i].m_PackType == PackTypeVideo)
//				continue;
			AppendPack(&SrcCell->m_Packs[i]);
		}
	}
	// We're done.
	assert(m_NumPacks <= m_NumAllocPacks);

if(ImageCell != SrcCell)
delete ImageCell;
	return 0;
}